/*
mod_maybe.c - Andrea Marchesini <baku@apache.org>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

mod_maybe.c -- Apache maybe module

LoadModule maybe_module modules/mod_maybe.so

AddOutputFilterByType MAYBE text/html
# it can be: 'off', 'on', 'maybe' or 'please'
MaybeEnabled on

#optional:
MaybeStatusCode 300

# it ca be: dev or stdlib
MaybeMethod dev
*/

#include <apr.h>
#include <httpd.h>
#include <http_log.h>
#include <http_core.h>
#include <http_config.h>
#include <http_request.h>
#include <http_protocol.h>
#include <ap_config.h>

#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>

#define MAYBE_NAME	"MAYBE"
#define MAYBE_SIZE	32

module AP_MODULE_DECLARE_DATA maybe_module;

typedef enum
{
  MAYBE_TYPE_OFF = 0,
  MAYBE_TYPE_ON,
  MAYBE_TYPE_PLEASE
    /* FIXME: something else... */
} maybe_type;

typedef enum
{
  MAYBE_METHOD_STDLIB = 0,
  MAYBE_METHOD_DEV,
  /* FIXME: something else... */
} maybe_method;

typedef struct
{
  maybe_type type;
  maybe_method method;

  int statuscode;

  int data[MAYBE_SIZE];
  int data_id;

} maybe_config;

static void
maybe_generate_stdlib (maybe_config * c)
{
  int i, mod;

  srandom (time (NULL));

  if (c->type == MAYBE_TYPE_PLEASE)
    mod = 10;
  else
    mod = 2;

  for (i = 0; i < MAYBE_SIZE; i++)
    c->data[i] = random () % mod;
}

static void
maybe_generate_dev (maybe_config * c)
{
  char *devices[] = { "/dev/random", "/dev/srandom", "/dev/urandom", NULL };
  char **p;
  int mod;

  if (c->type == MAYBE_TYPE_PLEASE)
    mod = 10;
  else
    mod = 2;

  for (p = devices; *p; p++)
    {
      int i, fd = open (*p, O_RDONLY | O_NDELAY);

      if (fd >= 0 && read (fd, c->data, sizeof (c->data)) == sizeof (c->data))
	{
	  close (fd);

	  for (i = 0; i < MAYBE_SIZE; i++)
	    c->data[i] = c->data[i] % mod;

	  return;
	}

      close (fd);
    }

  /* Fallback... */
  maybe_generate_stdlib (c);
}

static void
maybe_generate (maybe_config * c)
{
  switch (c->method)
    {
    case MAYBE_METHOD_STDLIB:
      maybe_generate_stdlib (c);
      break;

    case MAYBE_METHOD_DEV:
      maybe_generate_dev (c);
      break;
    }

  c->data_id = 0;
}

static int
maybe_boolean (maybe_config * c)
{
  int value;

  if (c->data_id == -1 || c->data_id == MAYBE_SIZE)
    maybe_generate (c);

  value = c->data[c->data_id];
  c->data_id++;

  return value;
}

static int
maybe_fixups (request_rec * r)
{
  maybe_config *c = ap_get_module_config (r->per_dir_config, &maybe_module);

  if (!c)
    return DECLINED;

  if (r->main)
    return DECLINED;

  if (c->type == MAYBE_TYPE_OFF)
    return DECLINED;

  if (!maybe_boolean (c))
    {
      ap_log_rerror (APLOG_MARK, APLOG_ERR, 0, r,
		     "Maybe: the request %s is blocked", r->uri);
      return c->statuscode;
    }

  return DECLINED;
}

static void *
create_dir_config (apr_pool_t * p, char *dir)
{
  maybe_config *c = apr_pcalloc (p, sizeof (maybe_config));

  c->statuscode = 503;
  c->data_id = -1;

  return (void *) c;
}

static void *
merge_dir_config (apr_pool_t * p, void *basev, void *overridesv)
{
  maybe_config *c = apr_palloc (p, sizeof (maybe_config));
  maybe_config *base = (maybe_config *) basev;
  maybe_config *overrides = (maybe_config *) overridesv;

  c->type = overrides->type ? overrides->type : base->type;
  c->statuscode =
    overrides->statuscode ? overrides->statuscode : base->statuscode;
  return c;
}

static const char *
set_enabled (cmd_parms * cmd, void *mconfig, const char *arg)
{
  maybe_config *c = (maybe_config *) mconfig;

  if (!strcasecmp (arg, "ON"))
    c->type = MAYBE_TYPE_ON;

  else if (!strcasecmp (arg, "MAYBE"))
    {
      if (!maybe_boolean (c))
	c->type = MAYBE_TYPE_ON;
    }

  else if (!strcasecmp (arg, "PLEASE"))
    c->type = MAYBE_TYPE_PLEASE;

  return NULL;
}

static const char *
set_statuscode (cmd_parms * cmd, void *mconfig, const char *arg)
{
  maybe_config *c = (maybe_config *) mconfig;
  c->statuscode = atoi (arg);

  return NULL;
}

static const char *
set_method (cmd_parms * cmd, void *mconfig, const char *arg)
{
  maybe_config *c = (maybe_config *) mconfig;

  if (!strcasecmp (arg, "STDLIB"))
    c->method = MAYBE_METHOD_STDLIB;

  else if (!strcasecmp (arg, "DEV"))
    c->method = MAYBE_METHOD_DEV;

  return NULL;
}

static const command_rec cmds[] = {
  AP_INIT_TAKE1 ("MaybeEnabled", set_enabled, NULL, OR_FILEINFO,
		 "Enable mod_maybe."),
  AP_INIT_TAKE1 ("MaybeStatusCode", set_statuscode, NULL, OR_FILEINFO,
		 "Set a status code (default 503)."),
  AP_INIT_TAKE1 ("MaybeMethod", set_method, NULL, OR_FILEINFO,
		 "Set a method for the 'maybe' flag (STDLIB, DEV)."),
  {NULL}
};

static void
register_hooks (apr_pool_t * p)
{
  ap_hook_fixups (maybe_fixups, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA maybe_module = {
  STANDARD20_MODULE_STUFF,
  create_dir_config,		/* create per-dir    config structures */
  merge_dir_config,		/* merge  per-dir    config structures */
  NULL,				/* create per-server config structures */
  NULL,				/* merge  per-server config structures */
  cmds,				/* table of config file commands       */
  register_hooks		/* register hooks */
};
